// POCSAG pieces of this code is based on rpitx by F5OEO (https://github.com/F5OEO/rpitx)
// so here's its license:
/*MIT License
Copyright (c) 2016 Galen Alderson
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
Fork and modification for rpitx (c)(F5OEO 2018)
** 11.09.2019 : Added Numeric Pager support by cuddlycheetah (github.com/cuddlycheetah)
** 14.10.2019 : Added Repeating Transmission + Single Preamble Mode
**
** fall 2021  : Modifications to run on an Arduino Pro Mini by Pena
*/

#include <time.h>
#include "LowPower.h" // This right here: https://github.com/rocketscream/Low-Power
#include <EEPROM.h>
#include <SPI.h>
#include <LoRa.h>

// Settings and constants for POCSAG
#define SYNC 0x7CD215D8
#define IDLE 0x7A89C197
#define FRAME_SIZE 2
#define BATCH_SIZE 16
#define PREAMBLE_LENGTH 576
#define FLAG_ADDRESS 0x000000
#define FLAG_MESSAGE 0x100000
#define FLAG_TEXT_DATA 0x3
#define TEXT_BITS_PER_WORD 20
#define TEXT_BITS_PER_CHAR 7
#define CRC_BITS 10

// Functions borrowed from rpitx start here:
uint32_t crc(uint32_t inputMsg) {
  uint32_t denominator = 0b11101101001;  // Need to set this value first before shifting
  denominator = denominator << 20;       // because denominator would now be zero on one line
  uint32_t msg = inputMsg << CRC_BITS;
  for (int column = 0; column <= 20; column++) {
    int msgBit = (msg >> (30 - column)) & 1;
    if (msgBit != 0) msg ^= denominator;
    denominator >>= 1;
  }
  return msg & 0x3FF;
}

uint32_t parity(uint32_t x) {
  uint32_t p = 0;
  for (int i = 0; i < 32; i++) {
    p ^= (x & 1);
    x >>= 1;
  }
  return p;
}

uint32_t encodeCodeword(uint32_t msg) {
  uint32_t fullCRC = (msg << CRC_BITS) | crc(msg);
  uint32_t p = parity(fullCRC);
  return (fullCRC << 1) | p;
}

uint32_t encodeASCII(uint32_t initial_offset, char *str, uint32_t *out) {
  uint32_t numWordsWritten = 0;
  uint32_t currentWord = 0;
  uint32_t currentNumBits = 0;
  uint32_t wordPosition = initial_offset;
  while (*str != 0) {
    unsigned char c = *str;
    str++;
    for (int i = 0; i < TEXT_BITS_PER_CHAR; i++) {
      currentWord <<= 1;
      currentWord |= (c >> i) & 1;
      currentNumBits++;
      if (currentNumBits == TEXT_BITS_PER_WORD) {
        *out = encodeCodeword(currentWord | FLAG_MESSAGE);
        out++;
        currentWord = 0;
        currentNumBits = 0;
        numWordsWritten++;
        wordPosition++;
        if (wordPosition == BATCH_SIZE) {
          *out = SYNC;
          out++;
          numWordsWritten++;
          wordPosition = 0;
        }
      }
    }
  }
  if (currentNumBits > 0)
  {
    currentWord <<= 20 - currentNumBits;
    *out = encodeCodeword(currentWord | FLAG_MESSAGE);
    out++;
    numWordsWritten++;
    wordPosition++;
    if (wordPosition == BATCH_SIZE) {
      *out = SYNC;
      out++;
      numWordsWritten++;
      wordPosition = 0;
    }
  }
  return numWordsWritten;
}

int addressOffset(long address) {
  return (address & 0x7) * FRAME_SIZE;
}

void encodeTransmission(int repeatIndex, long address, int fb, char *message, uint32_t *out) {
  if (repeatIndex == 0)
    for (int i = 0; i < PREAMBLE_LENGTH / 32; i++) {
      *out = 0xAAAAAAAA;
      out++;
    }
  uint32_t *start = out;
  *out = SYNC;
  out++;
  int prefixLength = addressOffset(address);
  for (int i = 0; i < prefixLength; i++) {
    *out = IDLE;
    out++;
  }
  *out = encodeCodeword(((address >> 3) << 2) | fb);
  out++;
  out += encodeASCII(addressOffset(address) + 1, message, out);
  *out = IDLE;
  out++;
  size_t written = out - start;
  size_t padding = (BATCH_SIZE + 1) - written % (BATCH_SIZE + 1);
  for (size_t i = 0; i < padding; i++) {
    *out = IDLE;
    out++;
  }
}

size_t textMessageLength(int repeatIndex, long address, int numChars) {
  size_t numWords = 0;
  numWords += addressOffset(address);
  numWords++;
  numWords += (numChars * TEXT_BITS_PER_CHAR + (TEXT_BITS_PER_WORD - 1)) / TEXT_BITS_PER_WORD;
  numWords++;
  numWords += BATCH_SIZE - (numWords % BATCH_SIZE);
  numWords += numWords / BATCH_SIZE;
  if (repeatIndex == 0) numWords += PREAMBLE_LENGTH / 32;
  return numWords;
}
// Functions borrowed from rpitx end here.


int pogsacpin = 7;            // 7 = pin that outputs POGSAC signal to the pager
int sleepdetectpin = 3;       // 3 = pin that watches pager's 1 hz pulse for "power off detection"
#define SLEEPTIME 15000       // When 15000 ms = 15 seconds have passed, Arduino goes to sleep
unsigned long lastActionTime; // Related to above: millis() of last known pulse from pager
int pocsagkeepalive = 0;      // Arduino sends null POCSAG every x minutes to avoid "out of range"
int SetRate;
int pindelay;
int SetFunctionBits;
int REPEAT_COUNT;
bool SetInverted;
// Settings for LoRa
#define SCK 13
#define MISO 12
#define MOSI 11
#define SS 10
#define RST 5
#define DIO0 2
#define DIO1 9
// EEPROM memory "slots"
#define MEM_SETTINGSEXIST 0
#define MEM_SPREADINGFACTOR 1
#define MEM_BANDWIDTH 2
#define MEM_CODINGRATE 3
#define MEM_PREAMBLELENGTH_HIGH 4
#define MEM_PREAMBLELENGTH_LOW 5
#define MEM_SYNCWORD 6
#define MEM_LORACRC 7
#define MEM_INVERTIQ 8
#define MEM_LNAGAIN 9
#define MEM_FREQ1 10
#define MEM_FREQ2 11
#define MEM_FREQ3 12
#define MEM_FREQ4 13
#define MEM_TXPOWER 14
#define MEM_POCSAGKEEPALIVE 15

void setup() {
  // Initializing serial port and throwing something for debug purposes, *IF* serial is connected.
  if (Serial) Serial.begin(9600);
  if (Serial) Serial.println(F("Booting..."));
  // Setting pins
  pinMode(pogsacpin, OUTPUT);
  digitalWrite(pogsacpin, HIGH);
  pinMode(sleepdetectpin, INPUT_PULLUP);

  // POGSAC settings
  SetRate = 512;
  pindelay = (1000000 / SetRate) * 0.991;  // YMMV, mine was a bit too slow w/o multiplier
  SetFunctionBits = 2;
  REPEAT_COUNT = 1;
  SetInverted = true;

  // Initializing RFM95 LoRa module
  pinMode(DIO1, INPUT);
  LoRa.setPins(SS, RST, DIO0);
  LoRa.setSPIFrequency(1E6);
  
  if (!LoRa.begin(868E6)) {
    if (Serial) Serial.println(F("LoRa initialization failed!"));
  } else {
    if (Serial) Serial.println(F("LoRa initialized."));
  }

  if(EEPROM.read(MEM_SETTINGSEXIST)==66) {
    // Yeehaw, there's settings written in EEPROM to read them and print them on serial
    if (Serial) Serial.println(F("EEPROM has settings:"));
    // LoRa Frequency
    long lora_freq = ((long)EEPROM.read(MEM_FREQ1) << 24) +
                     ((long)EEPROM.read(MEM_FREQ2) << 16) +
                     ((long)EEPROM.read(MEM_FREQ3) << 8) +
                     (long)EEPROM.read(MEM_FREQ4);
    if (Serial) Serial.print(F("  Frequency: "));
    if (Serial) Serial.println(lora_freq);
    LoRa.setFrequency(lora_freq);
    // LoRa Spreading Factor
    int lora_sf = EEPROM.read(MEM_SPREADINGFACTOR);
    if(lora_sf >= 7 && lora_sf <= 12) { // SF 6 needs implicit header mode (???), ignore that
      LoRa.setSpreadingFactor((int)lora_sf);
      if (Serial) Serial.print(F("  Spreading Factor: "));
      if (Serial) Serial.println(lora_sf);
    } else {
      if (Serial) Serial.println(F("  Spreading Factor: 7 (default)"));
    }
    // LoRa Signal Bandwidth
    long sbw = 0;
    switch ((int)EEPROM.read(MEM_BANDWIDTH)) {
      case 1: sbw = 7.8E3; break;
      case 2: sbw = 10.4E3; break;
      case 3: sbw = 15.6E3; break;
      case 4: sbw = 20.8E3; break;
      case 5: sbw = 31.25E3; break;
      case 6: sbw = 41.7E3; break;
      case 7: sbw = 62.5E3; break;
      case 8: sbw = 125E3; break;
      case 9: sbw = 250E3; break;
      case 10: sbw = 500E3; break;
    }
    if (sbw>0) {
      LoRa.setSignalBandwidth(sbw);
      if (Serial) Serial.print(F("  Signal Bandwidth: "));
      if (Serial) Serial.println(sbw);
    } else {
      if (Serial) Serial.println(F("  Signal Bandwidth: 125E3 (default)"));
    }
    // LoRa Coding Rate
    int lora_cr = EEPROM.read(MEM_CODINGRATE);
    if(lora_cr >= 5 && lora_cr <= 8) { // 5-8 is valid
      LoRa.setCodingRate4((int)lora_cr);
      if (Serial) Serial.print(F("  Coding Rate: 4/"));
      if (Serial) Serial.println(lora_cr);
    } else {
      if (Serial) Serial.println(F("  Coding Rate: 4/5 (default)"));
    }
    // LoRa Preamble Length
    long lora_pal = ((long)EEPROM.read(MEM_PREAMBLELENGTH_HIGH) << 8) + (long)EEPROM.read(MEM_PREAMBLELENGTH_LOW);
    if(lora_pal >= 6 & lora_pal <= 65535) { // 6-65535 is valid
      LoRa.setPreambleLength((long)lora_pal);
      if (Serial) Serial.print(F("  Preamble Length: "));
      if (Serial) Serial.println(lora_pal);
    } else {
      if (Serial) Serial.println(F("  Preamble Length: 8 (default)"));
    }
    // LoRa Sync Word
    char lora_sw = EEPROM.read(MEM_SYNCWORD);
    LoRa.setSyncWord(lora_sw);
    Serial.print(F("  Sync word: 0x"));
    Serial.println(lora_sw, HEX);
    // LoRa CRC
    if(EEPROM.read(MEM_LORACRC)>0) {
      LoRa.enableCrc();
      if (Serial) Serial.println(F("  CRC is enabled."));
    } else {
      LoRa.disableCrc();
      if (Serial) Serial.println(F("  CRC is disabled."));
    }
    // LoRa Invert IQ
    if(EEPROM.read(MEM_INVERTIQ)>0) {
      LoRa.enableInvertIQ();
      if (Serial) Serial.println(F("  Inverted IQ is enabled."));
    } else {
      LoRa.disableInvertIQ();
      if (Serial) Serial.println(F("  Inverted IQ is disabled."));
    }
    // LoRa LNA Gain
    int lnagain = EEPROM.read(MEM_LNAGAIN);
    if(lnagain >= 0 && lnagain <= 6) { // Valid 1-6, 0=autogain
      LoRa.setGain((int)lnagain);
      if (Serial) Serial.print(F("  LNA Gain: "));
      if (Serial) Serial.println(lnagain);
    } else {
      if (Serial) Serial.println(F("  LNA Gain: 0 (default)"));
    }
    // LoRa TX power (for test message)
    int txpower = EEPROM.read(MEM_TXPOWER);
    if(txpower >= 2 && txpower <= 20) { // Valid 2-20, default 17
      LoRa.setTxPower((int)txpower);
      if (Serial) Serial.print(F("  TX Power: "));
      if (Serial) Serial.println(txpower);
    } else {
      if (Serial) Serial.println(F("  TX Power: 17 (default)"));
    }
    int pocsagkeepalive = EEPROM.read(MEM_POCSAGKEEPALIVE);
    if (Serial) Serial.print(F("  POCSAG Keepalive: "));
    if (Serial) Serial.println(pocsagkeepalive);
  } else {
    // There's no default settings found so let's write them
    if (Serial) Serial.print(F("EEPROM has no settings, writing default settings..."));
    // Writing default frequency
    long lora_freq = 868E6;
    EEPROM.write(MEM_FREQ1, (lora_freq >> 24) & 0xFF);
    EEPROM.write(MEM_FREQ2, (lora_freq >> 16) & 0xFF);
    EEPROM.write(MEM_FREQ3, (lora_freq >> 8) & 0xFF);
    EEPROM.write(MEM_FREQ4, lora_freq & 0xFF);
    // Writing default spreading factor
    EEPROM.write(MEM_SPREADINGFACTOR, 7);
    // Writing default signal bandwidth
    EEPROM.write(MEM_BANDWIDTH, 8);
    // Writing default coding rate
    EEPROM.write(MEM_CODINGRATE, 5);
    // Writing default preamble length
    EEPROM.write(MEM_PREAMBLELENGTH_HIGH, 8 >> 8);
    EEPROM.write(MEM_PREAMBLELENGTH_LOW, 8 & 0xFF);
    // Writing default sync word
    EEPROM.write(MEM_SYNCWORD, 0x12);
    // Writing CRC off by default
    EEPROM.write(MEM_LORACRC, 0);
    // Writing Invert IQ off by default
    EEPROM.write(MEM_INVERTIQ, 0);
    // Writing default LNA gain
    EEPROM.write(MEM_LNAGAIN, 0);
    // Writing default TX power
    EEPROM.write(MEM_TXPOWER, 17);
    // Writing default POCSAG keepalive value (minutes, disabled by default)
    EEPROM.write(MEM_POCSAGKEEPALIVE, 0);
    // Writing that settings are stored
    EEPROM.write(MEM_SETTINGSEXIST, 66);
  }
  lastActionTime = millis();
  if (Serial) Serial.println(F("Up and running, ready for messages."));
}

// When Arduino is put to sleep and eventually woken up, this runs
void wakeUp() {
  asm volatile ("jmp 0x7800"); // Easier to reset the thing and start with clean slate
}

// All the menu* stuff is related to the interactive menu thing on serial port. Tough to explain how it works.
byte menuspot = 0;
char menuinput[12];
byte menuinputc = 0;
bool menuinputon = false;
bool keepalivemessage = false;  // When "out of range" is needed to be avoided, this will be true

void loop() {
  delay(25);

  // Read commands from serial port
  if ((Serial) && (Serial.available())) {
    char recv = Serial.read();

    if(recv >= 64 && recv <= 90) recv = recv + 32; // If user enters uppercase letters -> lowercase
    
    if (recv == 0x1B && menuspot > 0) { // Hitting ESC...
      menuinputon = false;
      recv = 0;
      menuspot = 0;
      Serial.println(F("\r\nGot escape, cancelling."));
    }
    if (menuinputon==true) { // Shoving entered characters to array, counting them, handling backspace, etc.
      menuinput[menuinputc] = recv;
      menuinputc ++;
      if ((recv==8 || recv==127) && menuinputc > 1) { menuinputc --; menuinput[menuinputc] = 0; menuinputc --; }
      if (menuinputc>=9) menuinputon = false;
      Serial.print(recv);
    }

    if (menuspot==0) { // Home
      if (recv==13) {
        Serial.println(F("*** Motorola Advisor LoRa Mod Quick Help ***"));
        Serial.println(F("L - Read LoRa module details."));
        Serial.println(F("S - Change LoRa module settings."));
        Serial.println(F("R - Reset Arduino."));
        Serial.println(F("X - Send short test message."));
      }
      if (recv=='l') {
        Serial.println(F("*** LoRa Module Details ***"));
        Serial.print(F("RSSI:            "));
        Serial.println((int)LoRa.rssi());
        Serial.print(F("Freq error     : "));
        Serial.println(LoRa.packetFrequencyError());
        Serial.print(F("Available bytes: "));
        Serial.println(LoRa.available());
        Serial.println(F("Register dump:"));
        LoRa.dumpRegisters(Serial);
        Serial.println("");
      }
      if (recv=='r') {
        menuspot = 255;
        Serial.print(F("\r\nPress y to reset Arduino, any other to cancel: "));
      }
      if (recv=='s') {
        menuspot = 10;
        Serial.println(F("\r\n*** LoRa Module Settings ***"));
        Serial.println(F("f=Frequency, s=Spreading Factor, b=Signal Bandwidth, r=Coding Rate, w=Sync Word,"));
        Serial.println(F("p=Preamble Length, c=CRC, i=Invert IQ, g=LNA Gain, t=TX Power, k=POCSAG Keepalive"));
        Serial.print(F("Select [f/s/b/r/w/p/c/i/g/t/k]: "));
      }
      if (recv=='x') {
        Serial.println(F("Sending test message."));
        LoRa.beginPacket();
        LoRa.print("TEST");
        LoRa.endPacket();
      }
    } else if(menuspot==10) { // Settings menu
      Serial.print(recv);
      for(int i = 0; i < sizeof(menuinput); ++i) menuinput[i] = (char)0;
      menuinputc = 0;
      switch (recv) {
        case 'f':
          Serial.println(F("\r\nWarning: frequency will be set to any valid number whether module supports it or not."));
          Serial.print(F("Enter frequency: "));
          menuspot = 11;
          menuinputon = true;
          break;
        case 's':
          Serial.print(F("\r\nEnter spreading factor [7-12]: "));
          menuspot = 12;
          menuinputon = true;
          break;
        case 'b': 
          Serial.println(F("\r\na=7.8E3, b=10.4E3, c=15.6E3, d=20.8E3, e=31.25E3,"));
          Serial.println(F("f=41.7E3, g=62.5E3, h=125E3, i=250E3, j=500E3"));
          Serial.print(F("Enter signal bandwidth [a-j]: "));
          menuspot = 13;
          break;
        case 'r':
          Serial.print(F("\r\nEnter coding rate [5-8]: "));
          menuspot = 14;
          break;
        case 'p': 
          Serial.print(F("\r\nEnter preamble length [6-65535]: "));
          menuspot = 15;
          menuinputon = true;
          break;
        case 'w':
          Serial.print(F("\r\nEnter sync word: 0x"));
          menuspot = 16;
          menuinputon = true;
          break;
        case 'c':
          Serial.print(F("\r\nEnable CRC [y/n]: "));
          menuspot = 17;
          break;
        case 'i': 
          Serial.print(F("\r\nInvert IQ [y/n]: "));
          menuspot = 18;
          break;
        case 'g':
          Serial.print(F("\r\nEnter LNA gain [0-6]: "));
          menuspot = 19;
          break;
        case 't':
          Serial.print(F("\r\nEnter TX power [2-20]: "));
          menuspot = 20;
          menuinputon = true;
          break;
        case 'k':
          Serial.print(F("\r\nEnter POCSAG Keepalive [1-255, 0=disabled]: "));
          menuspot = 21;
          menuinputon = true;
          break;
        default:
          Serial.println(F("\r\nInvalid selection, cancelled."));
      }
    } else if(menuspot==11) { // Settings: frequency
      if (recv==13) {
        long lora_freq = atol(menuinput);
        if(lora_freq > 0) {
          Serial.print(F("\r\nSetting frequency to "));
          Serial.print(lora_freq);
          Serial.print(F("..."));
          EEPROM.update(MEM_FREQ1, (lora_freq >> 24) & 0xFF);
          EEPROM.update(MEM_FREQ2, (lora_freq >> 16) & 0xFF);
          EEPROM.update(MEM_FREQ3, (lora_freq >> 8) & 0xFF);
          EEPROM.update(MEM_FREQ4, lora_freq & 0xFF);
          LoRa.setFrequency(lora_freq);
          Serial.println(F(" Done."));
        } else {
          Serial.print(F("\r\nInvalid value: "));
          Serial.println(lora_freq);
          Serial.println(F("Cancelled."));
        }
        menuspot = 0;
      }
    } else if(menuspot==12) { // Settings: spreading factor
      if (recv==13) {
        int lora_sf = atoi(menuinput);
        if(lora_sf >= 7 && lora_sf <= 12) {
          Serial.print(F("\r\nSetting spreading factor to "));
          Serial.print(lora_sf);
          Serial.print(F("..."));
          EEPROM.update(MEM_SPREADINGFACTOR, lora_sf);
          LoRa.setSpreadingFactor(lora_sf);
          Serial.println(F(" Done."));
        } else {
          Serial.print(F("\r\nInvalid value: "));
          Serial.println(lora_sf);
          Serial.println(F("Cancelled."));
        }
        menuspot = 0;
      }
    } else if(menuspot==13) { // Settings: signal bandwidth
      long lora_sbw = 0;
      char mem_sbw = 0;
      switch (recv) {
        case 'a': lora_sbw = 7.8E3; mem_sbw = 1; break;
        case 'b': lora_sbw = 10.4E3; mem_sbw = 2; break;
        case 'c': lora_sbw = 15.6E3; mem_sbw = 3; break;
        case 'd': lora_sbw = 20.8E3; mem_sbw = 4; break;
        case 'e': lora_sbw = 31.25E3; mem_sbw = 5; break;
        case 'f': lora_sbw = 41.7E3; mem_sbw = 6; break;
        case 'g': lora_sbw = 62.5E3; mem_sbw = 7; break;
        case 'h': lora_sbw = 125E3; mem_sbw = 8; break;
        case 'i': lora_sbw = 250E3; mem_sbw = 9; break;
        case 'j': lora_sbw = 500E3; mem_sbw = 10; break;
      }
      if (lora_sbw>0) {
        Serial.print(F("\r\nSetting signal bandwidth to "));
        Serial.print(lora_sbw);
        Serial.print(F("..."));
        EEPROM.update(MEM_BANDWIDTH, mem_sbw);
        LoRa.setSignalBandwidth(lora_sbw);
        Serial.println(F(" Done."));
      } else {
        Serial.print(F("\r\nInvalid value: "));
        Serial.println(lora_sbw);
        Serial.println(F("Cancelled."));
      }
      menuspot = 0;
    } else if(menuspot==14) { // Settings: coding rate
      int lora_cr = (char)recv - '0';
      if (lora_cr >= 5 && lora_cr <= 8) {
        Serial.print(F("\r\nSetting coding rate to "));
        Serial.print(lora_cr);
        Serial.print(F("..."));
        EEPROM.update(MEM_CODINGRATE, (int)lora_cr);
        LoRa.setCodingRate4(lora_cr);
        Serial.println(F(" Done."));
      } else {
        Serial.print(F("\r\nInvalid value: "));
        Serial.println(lora_cr);
        Serial.println(F("Cancelled."));
      }
      menuspot = 0;
    } else if(menuspot==15) { // Settings: preamble length
      if (recv==13) {
        int lora_pal = atoi(menuinput);
        if(lora_pal >= 8 && lora_pal <= 65535) {
          Serial.print(F("\r\nSetting preamble length to "));
          Serial.print(lora_pal);
          Serial.print(F("..."));
          EEPROM.update(MEM_PREAMBLELENGTH_HIGH, (lora_pal >> 8) & 0xFF);
          EEPROM.update(MEM_PREAMBLELENGTH_LOW, lora_pal & 0xFF);            
          LoRa.setPreambleLength(lora_pal);
          Serial.println(F(" Done."));
        } else {
          Serial.print(F("\r\nInvalid value: "));
          Serial.println(lora_pal);
          Serial.println(F("Cancelled."));
        }
        menuspot = 0;
      }
    } else if(menuspot==16) { // Settings: sync word
      if (recv==13) {
        int lora_sw = strtol(menuinput, NULL, 16);
        if(lora_sw >= 0 && lora_sw <= 255) {
          Serial.print(F("\r\nSetting sync word to 0x"));
          Serial.print(lora_sw, HEX);
          Serial.print(F("..."));
          EEPROM.update(MEM_SYNCWORD, lora_sw);
          LoRa.setSyncWord(lora_sw);
          Serial.println(F(" Done."));
        } else {
          Serial.print(F("\r\nInvalid value: "));
          Serial.println(lora_sw);
          Serial.println(F("Cancelled."));
        }
        menuspot = 0;
      }
    } else if(menuspot==17) { // Settings: CRC enabled
      switch (recv) {
        case 'y':
          Serial.print(F("\r\nSetting CRC on..."));
          LoRa.enableCrc();
          EEPROM.update(MEM_LORACRC, 1);
          Serial.println(F(" Done."));
          break;
        case 'n':
          Serial.print(F("\r\nSetting CRC off..."));
          LoRa.disableCrc();
          EEPROM.update(MEM_LORACRC, 0);
          Serial.println(F(" Done."));
          break;
        default:
          Serial.print(F("\r\nInvalid value: "));
          Serial.println(recv);
          Serial.println(F("Cancelled."));
          break;
      }
      menuspot = 0;
    } else if(menuspot==18) { // Settings: Invert IQ
      switch (recv) {
        case 'y':
          Serial.print(F("\r\nSetting invert IQ on..."));
          LoRa.enableInvertIQ();
          EEPROM.update(MEM_INVERTIQ, 1);
          Serial.println(F(" Done."));
          break;
        case 'n':
          Serial.print(F("\r\nSetting invert IQ off..."));
          LoRa.disableInvertIQ();
          EEPROM.update(MEM_INVERTIQ, 0);
          Serial.println(F(" Done."));
          break;
        default: 
          Serial.print(F("\r\nInvalid value: "));
          Serial.println(recv);
          Serial.println(F("Cancelled."));
          break;
      }
      menuspot = 0;
    } else if(menuspot==19) { // Settings: LNA Gain
      int lora_lnagain = (char)recv - '0';
      if (lora_lnagain >= 0 && lora_lnagain <= 6) {
        Serial.print(F("\r\nSetting LNA gain to "));
        Serial.print(lora_lnagain);
        Serial.print(F("..."));
        EEPROM.update(MEM_LNAGAIN, lora_lnagain);
        LoRa.setGain(lora_lnagain);
        Serial.println(F(" Done."));
      } else {
        Serial.print(F("\r\nInvalid value: "));
        Serial.println(lora_lnagain);
        Serial.println(F("Cancelled."));
      }
      menuspot = 0;
    } else if(menuspot==20) { // Settings: TX Power
      if (recv==13) {
        int lora_txp = atoi(menuinput);
        if(lora_txp >= 2 && lora_txp <= 20) {
          Serial.print(F("\r\nSetting TX power to "));
          Serial.print(lora_txp);
          Serial.print(F("..."));
          EEPROM.update(MEM_TXPOWER, lora_txp);
          LoRa.setTxPower(lora_txp);
          Serial.println(F(" Done."));
        } else {
          Serial.print(F("\r\nInvalid value: "));
          Serial.println(lora_txp);
          Serial.println(F("Cancelled."));
        }
        menuspot = 0;
      }      
    } else if(menuspot==21) { // Settings: POCSAG Keepalive
      if (recv==13) {
        int lora_psk = atoi(menuinput);
        if(lora_psk >= 0 && lora_psk <= 255) {
          Serial.print(F("\r\nSetting POCSAG Keepalive to "));
          if(lora_psk==0) {
            Serial.print(F("disabled"));
          } else {
            Serial.print(lora_psk);
          }
          Serial.print(F("..."));
          EEPROM.update(MEM_POCSAGKEEPALIVE, lora_psk);
          pocsagkeepalive = lora_psk;
          Serial.println(F(" Done."));
        } else {
          Serial.print(F("\r\nInvalid value: "));
          Serial.println(lora_psk);
          Serial.println(F("Cancelled."));
        }
        menuspot = 0;
      }      
    } else if(menuspot==255) { // Confirmation of reset
      if (recv=='y') {
        Serial.println(F("\r\nResetting..."));
        delay(100);
        asm volatile ("jmp 0x7800");
      } else {
        Serial.println(F("\r\nCancelled."));
        menuspot = 0;
      }
    }
  }

  // If we can read pager's pulse, let's update lastActionTime with current millis
  if (digitalRead(sleepdetectpin)==LOW) { lastActionTime = millis(); }
  // If difference between current and last action millis is more than SLEEPTIME, go to sleep
  if (((millis() - lastActionTime) >= SLEEPTIME)) {
    // Entering sleep mode
    if (Serial) Serial.println(F("Going to sleep."));
    // Turning LoRa module off to save power
    LoRa.end();
    delay(100);
    // Using interrupt to wake up from sleep. Pager sets pin low @ 1 Hz if it turns/is on.
    attachInterrupt(digitalPinToInterrupt(3), wakeUp, LOW);
    // Interrupt attached, going to sleep.
    LowPower.powerDown(SLEEP_FOREVER, ADC_OFF, BOD_OFF);
    detachInterrupt(digitalPinToInterrupt(3));
    // Let's not fall through since LoRa module is off and won't be initialized again...
    asm volatile ("jmp 0x7800");
  }

  // If POCSAG keepalive is enabled, Arduino throws POCSAG messages to pager so that it thinks
  // it's in range. If Arduino won't run due to too low voltage, pager will show the antenna icons.
  // Obviously you need to program the pager to use the 'out of range' setting.
  if (pocsagkeepalive>0) {
    if (pocsagkeepalive < (millis() / 1000 / 60)) {
      if (Serial) Serial.print(F("POCSAG keepalive hit, sending null message."));
      keepalivemessage = true;
    }
  }

  // Checking if LoRa packets have arrived.
  int packetSize = LoRa.parsePacket();
  // If there's something or need to avoid "out of range"...
  if ((packetSize) || (keepalivemessage==true)) {
    char bufferi[180];
    int readbytes = 0;
    if(keepalivemessage==false) { // If we're not avoiding "out of range", let's just accept what we got
      readbytes = LoRa.readBytes(bufferi, 180);
    } else { // And here we'll fake received message to avoid "out of range"
      String s = "12345:NULL";
      readbytes = s.length();
      s.toCharArray(bufferi, readbytes);
    }
    if (Serial) Serial.print("\r\nReceived (");
    if (Serial) Serial.print(readbytes);
    if (Serial) Serial.print("b|");
    if (Serial) Serial.print(LoRa.packetRssi());
    if (Serial) Serial.print("): ");
    if (Serial) Serial.println(bufferi);
    // Plenty of improvement required below... No idea how to make it better.
    // Separating RIC and message from the buffer; reading RIC until we get colon
    int colonIndex = 0;
    char addr[8];
    for(int i = 0; i <= sizeof(addr); i++) { addr[i] = (char)0; }
    char message[160];
    for(int i = 0; i <= sizeof(message); i++) { message[i] = (char)0; }
    for(int i = 0; i <= readbytes; i++) {
      if(colonIndex>0) message[i - colonIndex - 1] = bufferi[i];
      if(bufferi[i] == ':') colonIndex = i;
      if((colonIndex==0) && (i < 8)) addr[i-1] = bufferi[i];
    }
    // Converting extracted char array to long (not int)
    long address = atol(addr);
    // Zero rest of the message
    for(int i = readbytes - 1; i <= sizeof(message); i++) { message[i] = (char)0; }
    // Reset buffer back to zero
    for(int i = 0; i < sizeof(bufferi); ++i) bufferi[i] = (char)0;
    // RIC is valid?
    if((address > 0) && (address < 9999999)) {
      if (Serial) Serial.print(F("Valid RIC: "));
      if (Serial) Serial.println(address);
      srand(pindelay);
      // Next lines + for loop is stolen & modified from the LoRa library examples.
      size_t completeLength = 0;
      uint32_t *completeTransmission = (uint32_t *)malloc(sizeof(uint32_t) * 0);
      for (int x = 0; x < REPEAT_COUNT; x++) {
        size_t messageLength = textMessageLength(0, address, strlen(message));
        uint32_t *transmission = (uint32_t *)malloc(sizeof(uint32_t) * messageLength);
        encodeTransmission(0, address, SetFunctionBits, message, transmission);
        size_t beforeLength = completeLength + 0;
        completeLength += messageLength;
        completeTransmission = (uint32_t *)realloc(completeTransmission, sizeof(uint32_t) * completeLength);
        for (size_t byteI = 0; byteI < messageLength; byteI++) {
          completeTransmission[beforeLength + byteI] = transmission[byteI];
        }
      }
      // Now we have raw POCSAG stuff in a variable, need to make the pin pulse them to the pager
      static uint32_t oldtime = micros();
      for (int i = 0; i < completeLength; i++) {
        if (!SetInverted) completeTransmission[i] = ~completeTransmission[i];
        for (int j = 31; j >= 0; j--) {
          int b = (completeTransmission[i] >> j) & 0x1;
          if(b==1) { digitalWrite(pogsacpin, LOW); } else { digitalWrite(pogsacpin, HIGH); }
          while (micros() - oldtime < pindelay) {}
          oldtime = micros();
        }
      }
      digitalWrite(pogsacpin, HIGH);
      if (Serial) Serial.println("POGSAC done, reset...");
      delay(500);
      // Resetting Pro Nano because running this makes Arduino corrupt(?) after 2-3 messages.
      asm volatile ("jmp 0x7800");
    }
  }
}
